@@ -1,5 +1,7 @@
+from __future__ import annotations
+
 import itertools
-from typing import TYPE_CHECKING, Hashable, Iterable, Optional, Sequence, Tuple, Union
+from typing import TYPE_CHECKING, Hashable, Iterable, Sequence
 
 import numpy as np
 import pandas as pd
@@ -7,26 +9,25 @@ import xarray as xr
 
 from .aggregations import Aggregation, _atleast_1d
 from .core import (
-    factorize_,
+    _convert_expected_groups_to_index,
+    _get_expected_groups,
     groupby_reduce,
     rechunk_for_blockwise as rechunk_array_for_blockwise,
     rechunk_for_cohorts as rechunk_array_for_cohorts,
-    reindex_,
 )
-from .xrutils import is_duck_dask_array, isnull
 
 if TYPE_CHECKING:
     from xarray import DataArray, Dataset, Resample
 
 
-def _get_input_core_dims(group_names, dim, ds, to_group):
+def _get_input_core_dims(group_names, dim, ds, grouper_dims):
     input_core_dims = [[], []]
     for g in group_names:
         if g in dim:
             continue
         if g in ds.dims:
             input_core_dims[0].extend([g])
-        if g in to_group.dims:
+        if g in grouper_dims:
             input_core_dims[1].extend([g])
     input_core_dims[0].extend(dim)
     input_core_dims[1].extend(dim)
@@ -48,11 +49,11 @@ def _restore_dim_order(result, obj, by):
 
 
 def xarray_reduce(
-    obj: Union["Dataset", "DataArray"],
-    *by: Union["DataArray", Iterable[str], Iterable["DataArray"]],
-    func: Union[str, Aggregation],
+    obj: Dataset | DataArray,
+    *by: DataArray | Iterable[str] | Iterable[DataArray],
+    func: str | Aggregation,
     expected_groups=None,
-    isbin: Union[bool, Sequence[bool]] = False,
+    isbin: bool | Sequence[bool] = False,
     sort: bool = True,
     dim: Hashable = None,
     split_out: int = 1,
@@ -60,8 +61,9 @@ def xarray_reduce(
     method: str = "map-reduce",
     engine: str = "flox",
     keep_attrs: bool = True,
-    skipna: Optional[bool] = None,
-    min_count: Optional[int] = None,
+    skipna: bool | None = None,
+    min_count: int | None = None,
+    reindex: bool | None = None,
     **finalize_kwargs,
 ):
     """GroupBy reduce operations on xarray objects using numpy-groupies
@@ -142,6 +144,12 @@ def xarray_reduce(
         fewer than min_count non-NA values are present the result will be
         NA. Only used if skipna is set to True or defaults to True for the
         array's dtype.
+    reindex : bool, optional
+        Whether to "reindex" the blockwise results to `expected_groups` (possibly automatically detected).
+        If True, the intermediate result of the blockwise groupby-reduction has a value for all expected groups,
+        and the final result is a simple reduction of those intermediates. In nearly all cases, this is a significant
+        boost in computation speed. For cases like time grouping, this may result in large intermediates relative to the
+        original block size. Avoid that by using method="cohorts". By default, it is turned off for arg reductions.
     **finalize_kwargs :
         kwargs passed to the finalize function, like ``ddof`` for var, std.
 
@@ -173,6 +181,13 @@ def xarray_reduce(
 
     if isinstance(isbin, bool):
         isbin = (isbin,) * len(by)
+    if expected_groups is None:
+        expected_groups = (None,) * len(by)
+    if isinstance(expected_groups, (np.ndarray, list)):  # TODO: test for list
+        if len(by) == 1:
+            expected_groups = (expected_groups,)
+        else:
+            raise ValueError("Needs better message.")
 
     if not sort:
         raise NotImplementedError
@@ -185,12 +200,9 @@ def xarray_reduce(
         if isinstance(b, str) and not isbin_ and b in obj.dims and b not in obj.indexes
     )
 
-    by: Tuple["DataArray"] = tuple(obj[g] if isinstance(g, str) else g for g in by)  # type: ignore
+    by: tuple[DataArray] = tuple(obj[g] if isinstance(g, str) else g for g in by)  # type: ignore
 
-    if len(by) > 1 and any(is_duck_dask_array(by_.data) for by_ in by):
-        raise NotImplementedError("Grouping by multiple variables will compute dask variables.")
-
-    grouper_dims = set(itertools.chain(*tuple(g.dims for g in by)))
+    grouper_dims = tuple(itertools.chain(*tuple(g.dims for g in by)))
 
     if isinstance(obj, xr.DataArray):
         ds = obj._to_temp_dataset()
@@ -213,7 +225,7 @@ def xarray_reduce(
     # in the case where dim is Ellipsis, and by.ndim < obj.ndim
     # then we also broadcast `by` to all `obj.dims`
     # TODO: avoid this broadcasting
-    exclude_dims = set(ds.dims) - grouper_dims
+    exclude_dims = set(ds.dims) - set(grouper_dims)
     if dim is not None:
         exclude_dims -= set(dim)
     ds, *by = xr.broadcast(ds, *by, exclude=exclude_dims)
@@ -231,7 +243,7 @@ def xarray_reduce(
         # reducing along a dimension along which groups do not vary
         # This is really just a normal reduction.
         # This is not right when binning so we exclude.
-        if skipna:
+        if skipna and isinstance(func, str):
             dsfunc = func[3:]
         else:
             dsfunc = func
@@ -245,43 +257,29 @@ def xarray_reduce(
     axis = tuple(range(-len(dim), 0))
     group_names = tuple(g.name if not binned else f"{g.name}_bins" for g, binned in zip(by, isbin))
 
-    if len(by) > 1:
-        group_idx, expected_groups, group_shape, _, _, _ = factorize_(
-            tuple(g.data for g in by),
-            axis,
-            expected_groups,
-            isbin,
-        )
-        to_group = xr.DataArray(group_idx, dims=dim, coords={d: by[0][d] for d in by[0].indexes})
-    else:
-        if expected_groups is None and isinstance(by[0].data, np.ndarray):
-            uniques = np.unique(by[0].data)
-            nans = isnull(uniques)
-            if nans.any():
-                uniques = uniques[~nans]
-            expected_groups = (uniques,)
-        if expected_groups is None:
-            raise NotImplementedError(
-                "Please provide expected_groups if not grouping by a numpy-backed DataArray"
-            )
-        if isinstance(expected_groups, np.ndarray):
-            expected_groups = (expected_groups,)
-        if isbin[0]:
-            if isinstance(expected_groups[0], int):
-                raise NotImplementedError(
-                    "Does not support binning into an integer number of bins yet."
-                )
-                #    factorized, bins = pd.cut(by[0], bins=expected_groups[0], retbins=True)
-                group_shape = (expected_groups[0],)
-            else:
-                group_shape = (len(expected_groups[0]) - 1,)
-        else:
-            group_shape = (len(expected_groups[0]),)
-        to_group = by[0]
+    group_shape = [None] * len(by)
+    expected_groups = list(expected_groups)
 
+    # Set expected_groups and convert to index since we need coords, sizes
+    # for output xarray objects
+    for idx, (b, expect, isbin_) in enumerate(zip(by, expected_groups, isbin)):
+        if isbin_ and isinstance(expect, int):
+            raise NotImplementedError(
+                "flox does not support binning into an integer number of bins yet."
+            )
+        if expect is None:
+            if isbin_:
+                raise ValueError(
+                    f"Please provided bin edges for group variable {idx} "
+                    f"named {group_names[idx]} in expected_groups."
+                )
+            expected_groups[idx] = _get_expected_groups(b.data, sort=sort, raise_if_dask=True)
+
+    expected_groups = _convert_expected_groups_to_index(expected_groups, isbin, sort=sort)
+    group_shape = tuple(len(e) for e in expected_groups)
     group_sizes = dict(zip(group_names, group_shape))
 
-    def wrapper(array, to_group, *, func, skipna, **kwargs):
+    def wrapper(array, *by, func, skipna, **kwargs):
         # Handle skipna here because I need to know dtype to make a good default choice.
         # We cannnot handle this easily for xarray Datasets in xarray_reduce
         if skipna and func in ["all", "any", "count"]:
@@ -291,25 +289,7 @@ def xarray_reduce(
             if "nan" not in func and func not in ["all", "any", "count"]:
                 func = f"nan{func}"
 
-        result, groups = groupby_reduce(array, to_group, func=func, **kwargs)
-        if len(by) > 1:
-            # all groups need not be present. reindex here
-            # TODO: add test
-            reindexed = reindex_(
-                result,
-                from_=groups,
-                to=np.arange(np.prod(group_shape)),
-                fill_value=fill_value,
-                axis=-1,
-            )
-            result = reindexed.reshape(result.shape[:-1] + group_shape)
-        else:
-            # TODO: migrate this to core.groupby_reduce
-            # index out NaN or NaT groups; these should be last
-            if np.any(isnull(groups)):
-                result = result[..., :-1]
-                groups = groups[:-1]
-
+        result, *groups = groupby_reduce(array, *by, func=func, **kwargs)
         return result
 
     # These data variables do not have any of the core dimension,
@@ -325,11 +305,14 @@ def xarray_reduce(
             if is_missing_dim:
                 missing_dim[k] = v
 
+    input_core_dims = _get_input_core_dims(group_names, dim, ds, grouper_dims)
+    input_core_dims += [input_core_dims[-1]] * (len(by) - 1)
+
     actual = xr.apply_ufunc(
         wrapper,
-        ds.drop_vars(tuple(missing_dim) + bad_dtypes),
-        to_group,
-        input_core_dims=_get_input_core_dims(group_names, dim, ds, to_group),
+        ds.drop_vars(tuple(missing_dim) + bad_dtypes).transpose(..., *grouper_dims),
+        *by,
+        input_core_dims=input_core_dims,
         # for xarray's test_groupby_duplicate_coordinate_labels
         exclude_dims=set(dim),
         output_core_dims=[group_names],
@@ -346,14 +329,9 @@ def xarray_reduce(
             "min_count": min_count,
             "skipna": skipna,
             "engine": engine,
-            # The following mess exists because for multiple `by`s I factorize eagerly
-            # here before passing it on; this means I have to handle the
-            # "binning by single by variable" case explicitly where the factorization
-            # happens later allowing `by` to  be a dask variable.
-            # Another annoyance is that for resampling expected_groups is "disconnected"
-            # from "by" so we need the isbin part of the condition
-            "expected_groups": expected_groups[0] if len(by) == 1 and isbin[0] else None,
-            "isbin": isbin[0] if len(by) == 1 else False,
+            "reindex": reindex,
+            "expected_groups": tuple(expected_groups),
+            "isbin": isbin,
             "finalize_kwargs": finalize_kwargs,
         },
     )
@@ -364,9 +342,10 @@ def xarray_reduce(
         if all(d not in ds[var].dims for d in dim):
             actual[var] = ds[var]
 
-    for name, expect, isbin_ in zip(group_names, expected_groups, isbin):
-        if isbin_:
-            expect = [pd.Interval(left, right) for left, right in zip(expect[:-1], expect[1:])]
+    for name, expect in zip(group_names, expected_groups):
+        # Can't remove this till xarray handles IntervalIndex
+        if isinstance(expect, pd.IntervalIndex):
+            expect = expect.to_numpy()
         if isinstance(actual, xr.Dataset) and name in actual:
             actual = actual.drop_vars(name)
         actual[name] = expect
@@ -405,11 +384,11 @@ def xarray_reduce(
 
 
 def rechunk_for_cohorts(
-    obj: Union["DataArray", "Dataset"],
+    obj: DataArray | Dataset,
     dim: str,
-    labels: "DataArray",
+    labels: DataArray,
     force_new_chunk_at,
-    chunksize: Optional[int] = None,
+    chunksize: int | None = None,
 ):
     """
     Rechunks array so that each new chunk contains groups that always occur together.
@@ -448,7 +427,7 @@ def rechunk_for_cohorts(
     )
 
 
-def rechunk_for_blockwise(obj: Union["DataArray", "Dataset"], dim: str, labels: "DataArray"):
+def rechunk_for_blockwise(obj: DataArray | Dataset, dim: str, labels: DataArray):
     """
     Rechunks array so that group boundaries line up with chunk boundaries, allowing
     embarassingly parallel group reductions.
@@ -496,8 +475,8 @@ def _rechunk(func, obj, dim, labels, **kwargs):
 
 
 def resample_reduce(
-    resampler: "Resample",
-    func: Union[str, Aggregation],
+    resampler: Resample,
+    func: str | Aggregation,
     keep_attrs: bool = True,
     **kwargs,
 ):
@@ -521,11 +500,11 @@ def resample_reduce(
             by,
             func=func,
             method="blockwise",
-            expected_groups=(resampler._unique_coord.data,),
             keep_attrs=keep_attrs,
             **kwargs,
         )
         .rename({"__resample_dim__": dim})
         .transpose(dim, ...)
     )
+    result[dim] = resampler._unique_coord.data
     return result
