@@ -1,9 +1,14 @@
+from functools import reduce
+
 import numpy as np
+import pandas as pd
 import pytest
 from numpy_groupies.aggregate_numpy import aggregate
 
 from flox.core import (
+    _convert_expected_groups_to_index,
     _get_optimal_chunks_for_groups,
+    factorize_,
     find_group_cohorts,
     groupby_reduce,
     rechunk_for_cohorts,
@@ -111,7 +116,7 @@ def test_groupby_reduce(
     elif func == "count":
         expected = np.array(expected, dtype=int)
 
-    result, _ = groupby_reduce(
+    result, groups, = groupby_reduce(
         array,
         by,
         func=func,
@@ -120,13 +125,11 @@ def test_groupby_reduce(
         split_out=split_out,
         engine=engine,
     )
+    assert_equal(groups, [0, 1, 2])
     assert_equal(expected, result)
 
 
-@pytest.mark.parametrize("size", ((12,), (12, 5)))
-@pytest.mark.parametrize("func", ALL_FUNCS)
-def test_groupby_reduce_all(size, func, engine):
-
+def gen_array_by(size, func):
     by = np.ones(size[-1])
     rng = np.random.default_rng(12345)
     array = rng.random(size)
@@ -134,20 +137,57 @@ def test_groupby_reduce_all(size, func, engine):
         array[[1, 4, 5], ...] = np.nan
     elif "nanarg" in func and len(size) > 1:
         array[[1, 4, 5], 1] = np.nan
-
     if func in ["any", "all"]:
         array = array > 0.5
+    return array, by
 
-    finalize_kwargs = tuple({})
+
+@pytest.mark.parametrize("chunks", [None, 3, 4])
+@pytest.mark.parametrize("nby", [1, 2, 3])
+@pytest.mark.parametrize("size", ((12,), (12, 9)))
+@pytest.mark.parametrize("add_nan_by", [True, False])
+@pytest.mark.parametrize("func", ALL_FUNCS)
+def test_groupby_reduce_all(nby, size, chunks, func, add_nan_by, engine):
+    if chunks is not None and not has_dask:
+        pytest.skip()
+    if "arg" in func and engine == "flox":
+        pytest.skip()
+
+    array, by = gen_array_by(size, func)
+    if chunks:
+        array = dask.array.from_array(array, chunks=chunks)
+    by = (by,) * nby
+    by = [b + idx for idx, b in enumerate(by)]
+    if add_nan_by:
+        for idx in range(nby):
+            by[idx][2 * idx : 2 * idx + 3] = np.nan
+    by = tuple(by)
+    nanmask = reduce(np.logical_or, (np.isnan(b) for b in by))
+
+    finalize_kwargs = [{}]
     if "var" in func or "std" in func:
-        finalize_kwargs = finalize_kwargs + ({"ddof": 1}, {"ddof": 0})
+        finalize_kwargs = finalize_kwargs + [{"ddof": 1}, {"ddof": 0}]
+        fill_value = np.nan
+    else:
+        fill_value = None
 
     for kwargs in finalize_kwargs:
+        flox_kwargs = dict(func=func, engine=engine, finalize_kwargs=kwargs, fill_value=fill_value)
         with np.errstate(invalid="ignore", divide="ignore"):
-            expected = getattr(np, func)(array, axis=-1, **kwargs)
-        expected = np.expand_dims(expected, -1)
+            if "arg" in func and add_nan_by:
+                array[..., nanmask] = np.nan
+                expected = getattr(np, "nan" + func)(array, axis=-1, **kwargs)
+            else:
+                expected = getattr(np, func)(array[..., ~nanmask], axis=-1, **kwargs)
+        for _ in range(nby):
+            expected = np.expand_dims(expected, -1)
 
-        actual, _ = groupby_reduce(array, by, func=func, engine=engine, finalize_kwargs=kwargs)
+        actual, *groups = groupby_reduce(array, *by, **flox_kwargs)
+        assert actual.ndim == (array.ndim + nby - 1)
+        assert expected.ndim == (array.ndim + nby - 1)
+        expected_groups = tuple(np.array([idx + 1.0]) for idx in range(nby))
+        for actual_group, expect in zip(groups, expected_groups):
+            assert_equal(actual_group, expect)
         if "arg" in func:
             assert actual.dtype.kind == "i"
         assert_equal(actual, expected)
@@ -155,14 +195,11 @@ def test_groupby_reduce_all(size, func, engine):
         if not has_dask:
             continue
         for method in ["map-reduce", "cohorts", "split-reduce"]:
-            actual, _ = groupby_reduce(
-                da.from_array(array, chunks=3),
-                by,
-                func=func,
-                method=method,
-                engine=engine,
-                finalize_kwargs=kwargs,
-            )
+            if "arg" in func and method != "map-reduce":
+                continue
+            actual, *groups = groupby_reduce(array, *by, method=method, **flox_kwargs)
+            for actual_group, expect in zip(groups, expected_groups):
+                assert_equal(actual_group, expect)
             if "arg" in func:
                 assert actual.dtype.kind == "i"
             assert_equal(actual, expected)
@@ -224,7 +261,7 @@ def test_numpy_reduce_nd_md():
 
     expected = aggregate(by.ravel(), array.ravel(), func="sum")
     result, groups = groupby_reduce(array, by, func="sum", fill_value=123)
-    actual = reindex_(result, groups, np.unique(by), axis=0, fill_value=0)
+    actual = reindex_(result, groups, pd.Index(np.unique(by)), axis=0, fill_value=0)
     np.testing.assert_equal(expected, actual)
 
     array = np.ones((4, 2, 12))
@@ -232,18 +269,26 @@ def test_numpy_reduce_nd_md():
 
     expected = aggregate(by.ravel(), array.reshape(4, 24), func="sum", axis=-1, fill_value=0)
     result, groups = groupby_reduce(array, by, func="sum")
-    actual = reindex_(result, groups, np.unique(by), axis=-1, fill_value=0)
+    actual = reindex_(result, groups, pd.Index(np.unique(by)), axis=-1, fill_value=0)
     assert_equal(expected, actual)
 
     array = np.ones((4, 2, 12))
     by = np.broadcast_to(np.array([labels] * 2), array.shape)
     expected = aggregate(by.ravel(), array.ravel(), func="sum", axis=-1)
     result, groups = groupby_reduce(array, by, func="sum")
-    actual = reindex_(result, groups, np.unique(by), axis=-1, fill_value=0)
+    actual = reindex_(result, groups, pd.Index(np.unique(by)), axis=-1, fill_value=0)
     assert_equal(expected, actual)
 
+    array = np.ones((2, 3, 4))
+    by = np.ones((2, 3, 4))
+
+    actual, _ = groupby_reduce(array, by, axis=(1, 2), func="sum")
+    expected = np.sum(array, axis=(1, 2), keepdims=True).squeeze(2)
+    assert_equal(actual, expected)
+
 
 @requires_dask
+@pytest.mark.parametrize("reindex", [None, False, True])
 @pytest.mark.parametrize("func", ALL_FUNCS)
 @pytest.mark.parametrize("add_nan", [False, True])
 @pytest.mark.parametrize("dtype", (float,))
@@ -257,7 +302,7 @@ def test_numpy_reduce_nd_md():
         ((10, 12), (3, 3), 3),  # form 3
     ],
 )
-def test_groupby_agg_dask(func, shape, array_chunks, group_chunks, add_nan, dtype, engine):
+def test_groupby_agg_dask(func, shape, array_chunks, group_chunks, add_nan, dtype, engine, reindex):
     """Tests groupby_reduce with dask arrays against groupby_reduce with numpy arrays"""
 
     rng = np.random.default_rng(12345)
@@ -267,7 +312,7 @@ def test_groupby_agg_dask(func, shape, array_chunks, group_chunks, add_nan, dtyp
     if func in ["first", "last"]:
         pytest.skip()
 
-    if "arg" in func and engine == "flox":
+    if "arg" in func and (engine == "flox" or reindex):
         pytest.skip()
 
     labels = np.array([0, 0, 2, 2, 2, 1, 1, 2, 2, 1, 1, 0])
@@ -276,7 +321,9 @@ def test_groupby_agg_dask(func, shape, array_chunks, group_chunks, add_nan, dtyp
         labels[:3] = np.nan  # entire block is NaN when group_chunks=3
         labels[-2:] = np.nan
 
-    kwargs = dict(func=func, expected_groups=[0, 1, 2], fill_value=123)
+    kwargs = dict(
+        func=func, expected_groups=[0, 1, 2], fill_value=False if func in ["all", "any"] else 123
+    )
 
     expected, _ = groupby_reduce(array.compute(), labels, engine="numpy", **kwargs)
     actual, _ = groupby_reduce(array.compute(), labels, engine=engine, **kwargs)
@@ -293,38 +340,45 @@ def test_groupby_agg_dask(func, shape, array_chunks, group_chunks, add_nan, dtyp
 
     kwargs["expected_groups"] = [0, 2, 1]
     with raise_if_dask_computes():
-        actual, groups = groupby_reduce(array, by, engine=engine, **kwargs)
+        actual, groups = groupby_reduce(array, by, engine=engine, **kwargs, sort=False)
     assert_equal(groups, [0, 2, 1])
     assert_equal(expected, actual[..., [0, 2, 1]])
 
+    kwargs["expected_groups"] = [0, 2, 1]
+    with raise_if_dask_computes():
+        actual, groups = groupby_reduce(array, by, engine=engine, **kwargs, sort=True)
+    assert_equal(groups, [0, 1, 2])
+    assert_equal(expected, actual)
+
 
 def test_numpy_reduce_axis_subset(engine):
     # TODO: add NaNs
     by = labels2d
     array = np.ones_like(by)
-    result, _ = groupby_reduce(array, by, "count", axis=1, engine=engine)
+    kwargs = dict(func="count", engine=engine, fill_value=0)
+    result, _ = groupby_reduce(array, by, **kwargs, axis=1)
     assert_equal(result, [[2, 3], [2, 3]])
 
     by = np.broadcast_to(labels2d, (3, *labels2d.shape))
     array = np.ones_like(by)
-    result, _ = groupby_reduce(array, by, "count", axis=1, engine=engine)
+    result, _ = groupby_reduce(array, by, **kwargs, axis=1)
     subarr = np.array([[1, 1], [1, 1], [0, 2], [1, 1], [1, 1]])
     expected = np.tile(subarr, (3, 1, 1))
     assert_equal(result, expected)
 
-    result, _ = groupby_reduce(array, by, "count", axis=2, engine=engine)
+    result, _ = groupby_reduce(array, by, **kwargs, axis=2)
     subarr = np.array([[2, 3], [2, 3]])
     expected = np.tile(subarr, (3, 1, 1))
     assert_equal(result, expected)
 
-    result, _ = groupby_reduce(array, by, "count", axis=(1, 2), engine=engine)
+    result, _ = groupby_reduce(array, by, **kwargs, axis=(1, 2))
     expected = np.array([[4, 6], [4, 6], [4, 6]])
     assert_equal(result, expected)
 
-    result, _ = groupby_reduce(array, by, "count", axis=(2, 1), engine=engine)
+    result, _ = groupby_reduce(array, by, **kwargs, axis=(2, 1))
     assert_equal(result, expected)
 
-    result, _ = groupby_reduce(array, by[0, ...], "count", axis=(1, 2), engine=engine)
+    result, _ = groupby_reduce(array, by[0, ...], **kwargs, axis=(1, 2))
     expected = np.array([[4, 6], [4, 6], [4, 6]])
     assert_equal(result, expected)
 
@@ -338,7 +392,7 @@ def test_dask_reduce_axis_subset():
         result, _ = groupby_reduce(
             da.from_array(array, chunks=(2, 3)),
             da.from_array(by, chunks=(2, 2)),
-            "count",
+            func="count",
             axis=1,
             expected_groups=[0, 2],
         )
@@ -352,7 +406,7 @@ def test_dask_reduce_axis_subset():
         result, _ = groupby_reduce(
             da.from_array(array, chunks=(1, 2, 3)),
             da.from_array(by, chunks=(2, 2, 2)),
-            "count",
+            func="count",
             axis=1,
             expected_groups=[0, 2],
             fill_value=123,
@@ -365,7 +419,7 @@ def test_dask_reduce_axis_subset():
         result, _ = groupby_reduce(
             da.from_array(array, chunks=(1, 2, 3)),
             da.from_array(by, chunks=(2, 2, 2)),
-            "count",
+            func="count",
             axis=2,
             expected_groups=[0, 2],
         )
@@ -375,7 +429,7 @@ def test_dask_reduce_axis_subset():
         groupby_reduce(
             da.from_array(array, chunks=(1, 3, 2)),
             da.from_array(by, chunks=(2, 2, 2)),
-            "count",
+            func="count",
             axis=2,
         )
 
@@ -443,7 +497,7 @@ def test_groupby_reduce_nans(chunks, axis, groups, expected_shape, engine):
     result, _ = groupby_reduce(
         _maybe_chunk(array),
         _maybe_chunk(by),
-        "count",
+        func="count",
         expected_groups=groups,
         axis=axis,
         fill_value=0,
@@ -481,13 +535,24 @@ def test_groupby_all_nan_blocks(engine):
     assert_equal(actual, expected)
 
 
-def test_reindex():
-    array = np.array([1, 2])
-    groups = np.array(["a", "b"])
-    expected_groups = ["a", "b", "c"]
+@pytest.mark.parametrize("axis", (0, 1, 2, -1))
+def test_reindex(axis):
+    shape = [2, 2, 2]
     fill_value = 0
-    result = reindex_(array, groups, expected_groups, fill_value, axis=-1)
-    assert_equal(result, np.array([1, 2, 0]))
+
+    array = np.broadcast_to(np.array([1, 2]), shape)
+    groups = np.array(["a", "b"])
+    expected_groups = pd.Index(["a", "b", "c"])
+    actual = reindex_(array, groups, expected_groups, fill_value=fill_value, axis=axis)
+
+    if axis < 0:
+        axis = array.ndim + axis
+    result_shape = tuple(len(expected_groups) if ax == axis else s for ax, s in enumerate(shape))
+    slicer = tuple(slice(None, s) for s in shape)
+    expected = np.full(result_shape, fill_value)
+    expected[slicer] = array
+
+    assert_equal(actual, expected)
 
 
 @pytest.mark.xfail
@@ -541,7 +606,8 @@ def test_groupby_bins(chunk_labels, chunks, engine) -> None:
             engine=engine,
         )
     expected = np.array([3, 1, 0])
-    assert_equal(groups, np.array([0, 1, 2]))
+    for left, right in zip(groups, pd.IntervalIndex.from_arrays([1, 2, 4], [2, 4, 5]).to_numpy()):
+        assert left == right
     assert_equal(actual, expected)
 
 
@@ -610,13 +676,16 @@ def test_rechunk_for_cohorts(chunk_at, expected):
     assert rechunked.chunks == expected
 
 
+@pytest.mark.parametrize("chunks", [None, 3])
 @pytest.mark.parametrize("fill_value", [123, np.nan])
 @pytest.mark.parametrize("func", ALL_FUNCS)
-def test_fill_value_behaviour(func, fill_value, engine):
+def test_fill_value_behaviour(func, chunks, fill_value, engine):
     # fill_value = np.nan tests promotion of int counts to float
     # This is used by xarray
     if func in ["all", "any"] or "arg" in func:
         pytest.skip()
+    if chunks is not None and not has_dask:
+        pytest.skip()
 
     if func == "count":
 
@@ -629,12 +698,12 @@ def test_fill_value_behaviour(func, fill_value, engine):
 
     by = np.array([1, 2, 3, 1, 2, 3])
     array = np.array([np.nan, 1, 1, np.nan, 1, 1])
+    if chunks:
+        array = dask.array.from_array(array, chunks)
     actual, _ = groupby_reduce(
         array, by, func=func, engine=engine, fill_value=fill_value, expected_groups=[0, 1, 2, 3]
     )
-    expected = np.array(
-        [fill_value, npfunc([np.nan, np.nan]), npfunc([1.0, 1.0]), npfunc([1.0, 1.0])]
-    )
+    expected = np.array([fill_value, fill_value, npfunc([1.0, 1.0]), npfunc([1.0, 1.0])])
     assert_equal(actual, expected)
 
 
@@ -695,25 +764,24 @@ def test_cohorts_nd_by(func, method, axis, engine):
         pytest.xfail()
 
     kwargs = dict(func=func, engine=engine, method=method, axis=axis, fill_value=fill_value)
-    actual, _ = groupby_reduce(array, by, **kwargs)
+    actual, groups = groupby_reduce(array, by, **kwargs)
     expected, sorted_groups = groupby_reduce(array.compute(), by, **kwargs)
+    assert_equal(groups, sorted_groups)
     assert_equal(actual, expected)
 
     actual, groups = groupby_reduce(array, by, sort=False, **kwargs)
     if method == "cohorts":
         assert_equal(groups, [4, 3, 40, 2, 31, 1, 30])
+    elif method in ("split-reduce", "map-reduce"):
+        assert_equal(groups, [1, 30, 2, 31, 3, 4, 40])
     elif method == "blockwise":
         assert_equal(groups, [1, 30, 2, 31, 3, 40, 4])
-    else:
-        assert_equal(groups, [1, 2, 3, 4, 30, 31, 40])
-    reindexed = reindex_(actual, groups, sorted_groups)
+    reindexed = reindex_(actual, groups, pd.Index(sorted_groups))
     assert_equal(reindexed, expected)
 
 
 @pytest.mark.parametrize("func", ["sum", "count"])
-@pytest.mark.parametrize(
-    "fill_value, expected", ((None, np.floating), (0, np.integer), (np.nan, np.floating))
-)
+@pytest.mark.parametrize("fill_value, expected", ((0, np.integer), (np.nan, np.floating)))
 def test_dtype_promotion(func, fill_value, expected, engine):
     array = np.array([1, 1])
     by = [0, 1]
@@ -740,3 +808,18 @@ def test_empty_bins(func, engine):
     )
     expected = np.array([1.0, 1.0, np.nan])
     assert_equal(actual, expected)
+
+
+def test_datetime_binning():
+    time_bins = pd.date_range(start="2010-08-01", end="2010-08-15", freq="24H")
+    by = pd.date_range("2010-08-01", "2010-08-15", freq="15min")
+
+    (actual,) = _convert_expected_groups_to_index((time_bins,), isbin=(True,), sort=False)
+    expected = pd.IntervalIndex.from_arrays(time_bins[:-1], time_bins[1:])
+    assert_equal(actual, expected)
+
+    ret = factorize_((by.to_numpy(),), axis=0, expected_groups=(actual,))
+    group_idx = ret[0]
+    expected = pd.cut(by, time_bins).codes.copy()
+    expected[0] = 14  # factorize doesn't return -1 for nans
+    assert_equal(group_idx, expected)
