@@ -24,10 +24,11 @@ except ValueError:
     pass
 
 
+@pytest.mark.parametrize("reindex", [None, False, True])
 @pytest.mark.parametrize("min_count", [None, 1, 3])
 @pytest.mark.parametrize("add_nan", [True, False])
 @pytest.mark.parametrize("skipna", [True, False])
-def test_xarray_reduce(skipna, add_nan, min_count, engine):
+def test_xarray_reduce(skipna, add_nan, min_count, engine, reindex):
     arr = np.ones((4, 12))
 
     if add_nan:
@@ -47,7 +48,7 @@ def test_xarray_reduce(skipna, add_nan, min_count, engine):
 
     expected = da.groupby("labels").sum(skipna=skipna, min_count=min_count)
     actual = xarray_reduce(
-        da, "labels", func="sum", skipna=skipna, min_count=min_count, engine=engine
+        da, "labels", func="sum", skipna=skipna, min_count=min_count, engine=engine, reindex=reindex
     )
     assert_equal(expected, actual)
 
@@ -55,7 +56,13 @@ def test_xarray_reduce(skipna, add_nan, min_count, engine):
     da["labels2"][0] = np.nan
     expected = da.groupby("labels2").sum(skipna=skipna, min_count=min_count)
     actual = xarray_reduce(
-        da, "labels2", func="sum", skipna=skipna, min_count=min_count, engine=engine
+        da,
+        "labels2",
+        func="sum",
+        skipna=skipna,
+        min_count=min_count,
+        engine=engine,
+        reindex=reindex,
     )
     assert_equal(expected, actual)
 
@@ -66,9 +73,17 @@ def test_xarray_reduce(skipna, add_nan, min_count, engine):
     # assert_equal(expected, actual)
 
 
-def test_xarray_reduce_multiple_groupers(engine):
-    arr = np.ones((4, 12))
+# TODO: sort
+@pytest.mark.parametrize("pass_expected_groups", [True, False])
+@pytest.mark.parametrize("chunk", (True, False))
+def test_xarray_reduce_multiple_groupers(pass_expected_groups, chunk, engine):
+    if not has_dask and chunk:
+        pytest.skip()
 
+    if chunk and pass_expected_groups is False:
+        pytest.skip()
+
+    arr = np.ones((4, 12))
     labels = np.array(["a", "a", "c", "c", "c", "b", "b", "c", "c", "b", "b", "f"])
     labels = np.array(labels)
     labels2 = np.array([1, 2, 2, 1])
@@ -77,31 +92,39 @@ def test_xarray_reduce_multiple_groupers(engine):
         arr, dims=("x", "y"), coords={"labels2": ("x", labels2), "labels": ("y", labels)}
     ).expand_dims(z=4)
 
+    if chunk:
+        da = da.chunk({"x": 2, "z": 1})
+
     expected = xr.DataArray(
-        [[4, 4], [10, 10], [8, 8], [2, 2]],
+        [[4, 4], [8, 8], [10, 10], [2, 2]],
         dims=("labels", "labels2"),
-        coords={"labels": ["a", "c", "b", "f"], "labels2": [1, 2]},
+        coords={"labels": ["a", "b", "c", "f"], "labels2": [1, 2]},
     ).expand_dims(z=4)
 
-    actual = xarray_reduce(da, da.labels, da.labels2, func="count", engine=engine)
+    kwargs = dict(func="count", engine=engine)
+    if pass_expected_groups:
+        kwargs["expected_groups"] = (expected.labels.data, expected.labels2.data)
+
+    with raise_if_dask_computes():
+        actual = xarray_reduce(da, da.labels, da.labels2, **kwargs)
     xr.testing.assert_identical(expected, actual)
 
-    actual = xarray_reduce(da, "labels", da.labels2, func="count", engine=engine)
+    with raise_if_dask_computes():
+        actual = xarray_reduce(da, "labels", da.labels2, **kwargs)
     xr.testing.assert_identical(expected, actual)
 
-    actual = xarray_reduce(da, "labels", "labels2", func="count", engine=engine)
+    with raise_if_dask_computes():
+        actual = xarray_reduce(da, "labels", "labels2", **kwargs)
     xr.testing.assert_identical(expected, actual)
 
-    if has_dask:
-        with raise_if_dask_computes():
-            actual = xarray_reduce(
-                da.chunk({"x": 2, "z": 1}), da.labels, da.labels2, func="count", engine=engine
-            )
-        xr.testing.assert_identical(expected, actual)
 
-        with pytest.raises(NotImplementedError):
-            xarray_reduce(da.chunk({"x": 2, "z": 1}), "labels", "labels2", func="count")
-    # xr.testing.assert_identical(expected, actual)
+@requires_dask
+def test_dask_groupers_error():
+    da = xr.DataArray(
+        [1.0, 2.0], dims="x", coords={"labels": ("x", [1, 2]), "labels2": ("x", [1, 2])}
+    )
+    with pytest.raises(ValueError):
+        xarray_reduce(da.chunk({"x": 2, "z": 1}), "labels", "labels2", func="count")
 
 
 @requires_dask
@@ -158,7 +181,7 @@ def test_xarray_reduce_errors():
         xarray_reduce(da, by, func="mean", dim="foo")
 
     if has_dask:
-        with pytest.raises(NotImplementedError, match="provide expected_groups"):
+        with pytest.raises(ValueError, match="provide expected_groups"):
             xarray_reduce(da, by.chunk(), func="mean")
 
 
@@ -268,17 +291,16 @@ def test_xarray_groupby_bins(chunks, engine):
         array = array.chunk({"x": chunks})
         labels = labels.chunk({"x": chunks})
 
+    kwargs = dict(
+        dim="x",
+        func="count",
+        engine=engine,
+        expected_groups=np.array([1, 2, 4, 5]),
+        isbin=True,
+        fill_value=0,
+    )
     with raise_if_dask_computes():
-        actual = xarray_reduce(
-            array,
-            labels,
-            dim="x",
-            func="count",
-            engine=engine,
-            expected_groups=np.array([1, 2, 4, 5]),
-            isbin=True,
-            fill_value=0,
-        )
+        actual = xarray_reduce(array, labels, **kwargs)
     expected = xr.DataArray(
         np.array([3, 1, 0]),
         dims="labels_bins",
@@ -286,7 +308,18 @@ def test_xarray_groupby_bins(chunks, engine):
     )
     xr.testing.assert_equal(actual, expected)
 
-    # TODO: test cut_kwargs
+    # 3D array, 2D by, single dim, with NaNs in by
+    array = array.expand_dims(y=2, z=3)
+    labels = labels.expand_dims(y=2).copy()
+    labels.data[-1, -1] = np.nan
+    with raise_if_dask_computes():
+        actual = xarray_reduce(array, labels, **kwargs)
+    expected = xr.DataArray(
+        np.array([[[3, 1, 0]] * 3, [[3, 0, 0]] * 3]),
+        dims=("y", "z", "labels_bins"),
+        coords={"labels_bins": [pd.Interval(1, 2), pd.Interval(2, 4), pd.Interval(4, 5)]},
+    )
+    xr.testing.assert_equal(actual, expected)
 
 
 @requires_dask
@@ -303,3 +336,25 @@ def test_func_is_aggregation():
 
     with pytest.raises(ValueError):
         xarray_reduce(ds.Tair, ds.time.dt.month, func=mean, skipna=False)
+
+
+@requires_dask
+def test_cache():
+    pytest.importorskip("cachey")
+
+    from flox.cache import cache
+
+    ds = xr.Dataset(
+        {
+            "foo": (("x", "y"), dask.array.ones((10, 20), chunks=2)),
+            "bar": (("x", "y"), dask.array.ones((10, 20), chunks=2)),
+        },
+        coords={"labels": ("y", np.repeat([1, 2], 10))},
+    )
+
+    cache.clear()
+    xarray_reduce(ds, "labels", func="mean", method="cohorts")
+    assert len(cache.data) == 1
+
+    xarray_reduce(ds, "labels", func="mean", method="blockwise")
+    assert len(cache.data) == 2
